from pathlib import Path
from makru.store import Store
from .yamlh import yaml_load, yaml_dump
from .plug import PlugBox
import os
import sys
from typing import Dict, Tuple

class Panic(Exception):
    def __init__(self, message, *args) -> None:
        self.message = message
        super().__init__(message, *args)
    
    def __str__(self) -> str:
        return "panic: {}".format(self.message)

def panic(msg:str):
    raise Panic(msg)


def true_or_panic(test:bool, msg:str):
    if not test:
        panic(msg)


def file_exists(path, go_panic=False, msg="could not found {it}."):
    if not os.path.exists(path):
        if go_panic:
            panic(msg.format(it=path))
        else:
            return False
    return True


class Depend(object):
    def __init__(self, d):
        if not d.startswith("@"):
            parts = d.split("@")
            self.name = parts[0]
            self.dtype = "direct"
            if len(parts) > 1:
                self.version = parts[1]
            else:
                self.version = None
        else:
            # @(opt/path)
            self.dtype = "path"
            self.path = d[2:-1]

    def __repr__(self):
        if self.dtype == "direct":
            return "{} {}".format(self.name, self.version if hasattr(self, "version") else "*")
        elif self.dtype == "path":
            return "@({})".format(self.path)
        else:
            return "Depend@{}".format({"type": self.type})


class CompilingDesc(object):
    def __init__(self, sandbox: "Sandbox", **kvargs):
        self.sandbox = sandbox
        self.config = kvargs
        self.root = kvargs["root"]
        self.dependencies = list(map(self.make_depend_obj,kvargs.get("dependencies", [])))
        self.name = kvargs["name"]
        self.type = kvargs["type"]
        self.use = kvargs["use"]
        self.action_arg = kvargs.get("action_arg", None)
        self.action_default_arg = kvargs.get("action_default_arg", None)

    def make_depend_obj(self, s):
        return Depend(s)

    @property
    def gvars(self):
        return self.sandbox.gconfig['vars']
    
    def expand_path(self, path):
        return self.sandbox.expand_path(path)

class Sandbox(object):
    def __init__(self, confpath, **globalconf):
        self.plugbox = PlugBox()
        self.gconfig = globalconf
        self.confpath = os.path.realpath(confpath)
        file_exists(self.confpath, go_panic=True)
        self.root = os.path.split(os.path.realpath(confpath))[0]
        self.config = self.prepare_config(confpath)
        self.private_actions: Dict[str, function] = {}
        self._info_store: Store = None
        self._build_store: Store = None
    
    @property
    def info_store(self):
        if not self._info_store:
            self.init_info_store()
        return self._info_store
    
    def init_info_store(self):
        dir = str(Path(self.root) / ".makru")
        return self.make_store_under(dir, "info")
    
    @property
    def build_store(self):
        return self._build_store
    
    @build_store.setter
    def set_build_store(self, store: Store):
        if self._build_store:
            panic("build store could not be overrided, old: {}, new: {}".format(self._build_store, store))
        self._build_store = store
    
    @staticmethod
    def make_store_under(directory: str, name: str) -> Store:
        filename = "{}.sqlite3".format(name)
        dir_path = Path(directory)
        if not Path(directory).exists():
            dir_path.mkdir(parents=True, exist_ok=True)
        file_path = dir_path / filename
        return Store(str(file_path))
    
    def add_private_action(self, name: str, action=None):
        true_or_panic(
            name.startswith('_'), "private action's name should be leaded by '_': {}".format(name)
        )
        if action:
            self.private_actions[name] = action
        else:
            def wrapper(f):
                self.add_private_action(name, f)
                return f
            return wrapper

    @staticmethod
    def prepare_config(confpath):
        with open(confpath, 'r') as f:
            return yaml_load(f.read())

    def ready(self):
        self.plug_search()
        self.plugbox.runhook("on_sandbox_load", self)

    def plug_check_and_load(self, user_request:str, check="compile", panic_if_notfound=True):
        if self.plugbox.exists(user_request):
            plugin = self.plugbox.load(user_request)
            if hasattr(plugin, check):
                return plugin
            else:
                panic("plugin {} requested in {} does not support {} hook".format(user_request, self.confpath, check))
        else:
            if panic_if_notfound:
                panic("could not found plugin {} requested in {}, all plugins: {}, searched paths: {}".format(user_request, self.confpath, self.plugbox.plugin_names(), self.plugbox.searchpaths))

    def plug_search(self):
        if "plugin_paths" in self.gconfig:
            for p in self.gconfig["plugin_paths"]:
                self.plugbox.searchpaths.append(self.expand_path(p))
        if "plugin_paths" in self.config:
            for p in self.config["plugin_paths"]:
                self.plugbox.searchpaths.append(self.expand_path(p))
        self.plugbox.searchpaths.append(self.expand_path("./makru/plugins"))

    def build_desc(self, *, action_default_arg=None):
        config = self.config
        config['root'] = self.root
        return CompilingDesc(self, action_default_arg=action_default_arg, **config)

    def compile(self):
        target_plugin = self.plug_check_and_load(self.config["use"])
        target_plugin.compile(self.build_desc())

    def get_actions_conf(self):
        return self.config.get("actions", {})

    @staticmethod
    def action_split(action_name:str) -> Tuple[str, str]:
        parts = action_name.split('(')
        name = parts[0]
        if len(parts) < 2:
            return (name, None)
        darg = parts[1][:-1]
        return (name, darg)
    
    def is_action_exists(self, name):
        actions = self.get_actions_conf()
        target_action_s = actions.get(name, None)
        if target_action_s:
            parts = target_action_s.split(":")
            true_or_panic(
                len(parts) == 2,
                "action defined in wrong format, the right format is <plugin name>:<function name>, recviced is {}".format(target_action_s)
            )
            plugin_name = parts[0]
            function_name, default_arg = self.action_split(parts[1])
            plugin = self.plug_check_and_load(plugin_name, check=function_name, panic_if_notfound=False)
            if not plugin:
                return False
            return getattr(plugin, function_name, None) != None
        elif name in self.gconfig["bulitin_actions"]:
            return True
        elif name in self.private_actions:
            return True
        else:
            return False

    def run_action(self, name, arg=None):
        actions = self.get_actions_conf()
        target_action_s = actions.get(name, None)
        if target_action_s:
            parts = target_action_s.split(":")
            true_or_panic(
                len(parts) == 2,
                "action defined in wrong format, the right format is <plugin name>:<function name>, recviced is {}".format(target_action_s)
            )
            plugin_name = parts[0]
            function_name, default_arg = self.action_split(parts[1])
            plugin = self.plug_check_and_load(plugin_name, check=function_name)
            return getattr(plugin, function_name)(self.build_desc(action_default_arg=default_arg))
        elif name in self.gconfig["bulitin_actions"]:
            return self.gconfig["bulitin_actions"][name]()
        elif name in self.private_actions:
            return self.private_actions[name](self.build_desc())
        else:
            panic("could not find action {}".format(name))

    def expand_path(self, s):
        if s.startswith("./"):
            return os.path.join(self.root, s[2:])
        elif not s.startswith('/'):
            return os.path.join(self.root, s)
        else:
            return s


class Makru(object):
    def __init__(self, arguments=None):
        self._autorun_compile = True
        self.arguments = arguments if arguments else sys.argv[1:]
        self.options = self.scanflags()
        self.globalconfig = {
            "plugin_paths": [],
            "root_config_path": self.options.get("config_path", "./makru.yaml"),
            "bulitin_actions": {
                "compile": self.compile,
                "list_actions": self.list_actions,
            },
            "vars": self.options["vars"],
        }

    def main(self):
        if len(self.options["actions"]) > 0:
            for aname in self.options["actions"]:
                if "(" in aname and aname.endswith(')'):
                    parts = aname.split("(")
                    true_or_panic(
                        len(parts) == 2,
                        "found too more ( in your action call {}, could not parse it correctly".format(aname)
                    )
                    name = parts[0]
                    arg = parts[1][:-1]
                    self.action(name, arg)
                else:
                    self.action(aname)
        if self.options["do_compile"] or self._autorun_compile:
            self.compile()

    def build_sandbox(self):
        sandbox = Sandbox(self.globalconfig["root_config_path"], **self.globalconfig)
        sandbox.ready()
        return sandbox

    def compile(self):
        self.build_sandbox().compile()
    
    def list_actions(self):
        sandbox = self.build_sandbox()
        action_conf: Dict[str, str] = sandbox.get_actions_conf()
        print("All actions declared in \"{}\":".format(sandbox.confpath))
        for k in action_conf.keys():
            if sandbox.is_action_exists(k):
                print("  {} -> {}".format(k, action_conf[k]))
            else:
                print("  {} -> {} (not found)".format(k, action_conf[k]))
        if self.options['vars'].get('all', None):
            print("Private actions:")
            for k in sandbox.private_actions.keys():
                print("  {} -> {}".format(k, sandbox.private_actions[k]))

    def action(self, name, arg=None):
        sandbox = self.build_sandbox()
        if arg:
            sandbox.config["action_arg"] = arg
        result = sandbox.run_action(name, arg)
        if self.globalconfig['vars'].get('action-print-return', False):
            print(result)
        return result

    def scanflags(self):
        options = {
            "config_path": "./makru.yaml",
            "actions": [],
            "do_compile": False,
            "vars": {},
        }
        shell_like_action_arg = False
        for arg in self.arguments:
            if arg.startswith("-F"):
                options["config_path"] = self.correct_config_path(arg[2:])
            elif arg.startswith("--"):
                parts = arg[2:].split('=')
                if len(parts) == 1:
                    key = parts[0]
                    if key[0] == "!":
                        options['vars'][key[1:]] = False
                    else:
                        options['vars'][key] = True
                elif len(parts) == 2:
                    key = parts[0]
                    value = self.str2val(parts[1])
                    options['vars'][key] = value
                else:
                    panic("could not parse command line argument: undefined format {}".format(arg))
            elif arg.startswith(":"):
                options["actions"].append(arg[1:])
                self._autorun_compile = False
                shell_like_action_arg = True
            elif shell_like_action_arg and (not arg.startswith(":")) and (not arg.startswith("-")):
                a = options["actions"][-1]
                options["actions"][-1] = "{}({})".format(a, arg)
            elif arg == "-C":
                options["do_compile"] = True
            else:
                panic("could not parse command line argument: unknown option {}".format(arg))
            shell_like_actionn_arg = False
        return options

    @staticmethod
    def correct_config_path(path):
        if os.path.isdir(path):
            return os.path.join(path, "makru.yaml")
        else:
            return path

    @staticmethod
    def str2val(s):
        def isnum(s):
            number_strs = list(map(str, range(0,10)))
            is_num = False
            if len(s) == 1:
                return s in number_strs
            else:
                return (s[0] in number_strs) and isnum(s)
        if (s.startswith('""') and s.endswith('""')) or (s.startswith("''") and s.endswith("''")):
            return s[1:-1]
        if s == 'true':
            return True
        elif s == 'false':
            return False
        elif isnum(s):
            return int(s)
        else:
            return s



import os
import pytest
from . import helpers
from io import StringIO

class TestChooseFile:
    def test_can_return_path_if_file_exists(self, tmpdir):
        the_file_path = self.write_sample(tmpdir, 'test.test')

        result = helpers.choose_file([self.build_path_template(tmpdir)], "test.test")
        assert result
        assert the_file_path == result

    def test_can_return_none_if_file_does_not_exist(self, tmpdir):
        result = helpers.choose_file([self.build_path_template(tmpdir)], 'test')
        assert (not result)

    def test_can_return_correct_path_if_file_exists_when_have_multi_templates(self, tmpdir):
        the_file_path = self.write_sample(tmpdir, 'test.test')

        result = helpers.choose_file(['./?', self.build_path_template(tmpdir)], 'test.test')
        assert result
        assert the_file_path == result

    @staticmethod
    def build_path_template(path):
        return os.path.join(path, '?')

    @staticmethod
    def write_sample(path, filename):
        fpath = os.path.join(path, filename)
        with open(fpath, 'w+') as f:
            f.write("sample")
        return fpath


class TextRecorder(object):
    def __init__(self):
        self.buffer = StringIO()

    def record(self, x):
        self.buffer.write(x)
    
    @property
    def text(self):
        return self.buffer.getvalue()


class TestCheckBinary:
    def test_can_find_echo(self):
        """
        This test depends on the fact that most system have a binary "echo"
        """
        assert helpers.check_binary("echo")


class TestShell:
    def test_can_run_command(self):
        """
        This test depends on a fact that most system have a binary "echo"
        """
        if not helpers.check_binary("echo"):
            pytest.skip("could not found binary echo for this test")
        recorder = TextRecorder()

        helpers.shell(["echo", "test"], printf=recorder.record)

        lines = list(filter(lambda x: x != '', recorder.text.split(os.linesep)))
        assert len(lines) == 2
        assert lines[0].startswith('$')
        assert "echo test" in lines[0]
        assert lines[1] == "test"


class TestVersionHelpers:
    def test_compare_version_smoke_test(self):
        assert helpers.compare_version("0.1.1-alpha.2", "0.1.0-beta.10") > 0

    def test_match_version_smoke_test(self):
        assert helpers.match_version("0.1.2-alpha.2", ">0.1.0") == True
        assert helpers.match_version("0.2.1-pre.2+build.3", "<0.1.10") == False
    
    def test_is_semver_can_detect_non_semver_string_smoke_test(self):
        assert helpers.is_semver("20201210") == False
    
    def test_is_semver_can_detect_semver_smoke_test(self):
        assert helpers.is_semver("2.0.0-beta.2+build.4") == True
    
    def test_is_prerelease_smoke_test(self):
        assert helpers.is_prerelease("2.0.0") == False
        assert helpers.is_prerelease("2.0.0-beta.0") == True
    
    def test_pick_latest_release_can_pick_latest_release(self):
        DATA = [
            "0.2.0",
            "0.1.0-beta.6",
            "0.1.0-beta.10",
            "0.1.0",
            "0.2.0-alpha.4+build1",
            "0.3.0",
        ]
        assert helpers.pick_latest_release(DATA) == "0.3.0"
    
    def test_pick_max_version_can_pick_max_version(self):
        DATA = [
            "0.1.0-alpha.4",
            "0.1.0-beta.1",
            "0.1.0-beta.1+build.1",
            "0.2.0",
            "0.3.0-alpha.1",
            "0.3.0-alpha.2+build.1",
            "0.3.0-alpha.2+build.2",
        ]
        assert helpers.pick_max_version(DATA) == "0.3.0-alpha.2+build.2"



import sqlite3

class Store(object):
    def __init__(self, path) -> None:
        self.conn = sqlite3.connect(path)

    def get_table(self, name: str) -> 'Table':
        self.conn.execute("CREATE TABLE IF NOT EXISTS {} (key TEXT PRIMARY KEY, value TEXT)".format(name))
        self.conn.commit()
        return Table(self.conn, name)
    
    def close(self):
        self.conn.close()
    
    def remove_table(self, name:str):
        self.conn.execute("DROP TABLE IF EXISTS {};".format(name))

class Table(object):
    def __init__(self, conn: sqlite3.Connection, name: str) -> None:
        self.conn = conn
        self.name = name

    def set(self, key: str, value: str):
        if value == None:
            self.conn.execute("DELETE FROM {} WHERE key=?".format(self.name), (key,))
        else:
            self.conn.execute("REPLACE INTO {} (key, value) VALUES (?, ?)".format(self.name), (key, value))
        self.conn.commit()

    def get(self, key: str):
        cursor = self.conn.execute("SELECT value FROM {} WHERE key=?".format(self.name), (key,))
        value = cursor.fetchone()
        if value:
            return value[0]
        else:
            return None



from .store import Store, Table

class TestStore:
    def test_store_can_open_in_memory_sqlite3_database(self):
        store = Store(":memory:") # :memory: means open a in-memory database
        store.get_table("example")
    
    def test_store_can_remove_table(self):
        store = Store(":memory:")
        t = store.get_table("example")
        t.set("value1", "yah")
        assert t.get("value1") == "yah"
        store.remove_table("example")
        store.get_table("example")
        assert t.get("value1") == None
    
    def test_store_can_remove_table_even_when_table_does_not_exists(self):
        store = Store(":memory:")
        store.remove_table("this_table_does_not_exists")

def get_table(table_name):
    store = Store(":memory:")
    return store.get_table(table_name)

class TestTable:
    def test_table_could_set_values(self):
        t = get_table("example")
        t.set("value1", "ok")
        t.set("value2", "ok")
    
    def test_table_could_get_values(self):
        t = get_table("example")
        t.set("value1", "ok")
        assert t.get("value1") == "ok"
    
    def test_table_could_return_none_if_non_defined(self):
        t = get_table("example")
        assert t.get("unset_value") == None
    
    def test_table_remove_the_pair_when_value_sets_none(self):
        t = get_table("example")
        t.set("value1", "ok")
        assert t.get("value1") == "ok"
        t.set("value1", None)
        assert t.get("value1") == None



from pluginbase import PluginBase

class PlugBox(object):
    def __init__(self):
        self.pbase = PluginBase(
            package = "makru.plugins",
        )
        self.searchpaths = []
        self._source = None

    def source(self):
        if self._source:
            return self._source
        self._source = self.pbase.make_plugin_source(
            searchpath = self.searchpaths,
            persist = True,
        )
        return self._source

    def plugin_names(self):
        return self.source().list_plugins()

    def exists(self, name:str):
        return name in self.plugin_names()

    def load(self, name:str):
        return self.source().load_plugin(name)

    def runhook(self, name:str, *args, **kvargs):
        for n in self.plugin_names():
            p = self.load(n)
            if hasattr(p,name):
                getattr(p, name)(*args, **kvargs)



from setuptools import setup

def get_version():
    with open("VERSION", encoding='utf-8') as f:
        return f.read()

def get_requires():
    with open("requirements.txt", encoding='utf-8') as f:
        return f.read()

def get_readme():
    with open("README.md", encoding='utf-8') as f:
        return f.read()

def next_breakversion(s):
    n = int(s[0])
    return "{}.0.0".format(str(n+1))

def correctfmt(s):
    parts = s.split("==")
    name = parts[0]
    version = parts[1]
    limitop = ">={},<{}".format(version, next_breakversion(version))
    return "{}({})".format(name,limitop)

requirements = list(map(
    correctfmt,
    filter(lambda x: x != '', get_requires().split("\n"))
))

VERSION = get_version()

README = get_readme()

setup(
    name="makru",
    version=VERSION,
    description="A simple, readable way to compile your program",
    long_desciption=README,
    long_desciption_content_type="text/markdown",
    author="The Makru Contributors",
    url="https://gitlab.com/jinwa/makru",
    packages=['makru'],
    package_data={
        "makru": ["requirements.txt", "VERSION"]
    },
    entry_points="""
    [console_scripts]
    makru = makru.main:main
    """,
    install_requires=requirements
)



import pytest
from pathlib import Path
from . import plug

class TestPlugBox:
    def get_self_plugbox(self):
        p = Path('.') / 'tools' / 'plugins'
        if p.exists():
            box = plug.PlugBox()
            box.searchpaths.append(p)
            return box
    
    def test_can_create_plugbox(self):
        plug.PlugBox()
    
    def test_can_find_plugin_correctly(self):
        PLUGNAME = "empty-plugin"
        box = self.get_self_plugbox()
        if box:
            box.exists(PLUGNAME)
        else:
            pytest.skip("could not find correct plugin folder for test")
    
    def test_can_load_plugin_correctly(self):
        pytest.skip("this test always died unexpectly")
        # Looks like something (pytest?) modified the importing, makes pluginbase could not load module from internal space
        PLUGNAME = "empty-plugin"
        box = self.get_self_plugbox()
        if box:
            box.load(PLUGNAME)
        else:
            pytest.skip("could not find correct plugin folder for test")



import os
from pathlib import Path
import subprocess as subp
from hashlib import sha224
from typing import Iterable, Union
import semver

def exprog(args, cwd=None, env=None):
    env = env if env else os.environ
    cwd = cwd if cwd else os.getcwd()
    return subp.Popen(
        args,
        cwd=cwd,
        env=env,
        stdout=subp.PIPE,
        stderr=subp.PIPE,
        encoding="utf-8",
        universal_newlines=True
    )

def shell(args, cwd=None, env=None, printf=lambda x: print(x, end='')):
    printf(
        "$ {}\n".format(
            ' '.join(args) if not isinstance(args, str) else args
        )
    )
    with exprog(args, cwd=cwd, env=env) as p:
        while p.poll() == None:
            printf(p.stdout.read(1))
        if p.stdout.readable():
            printf(p.stdout.read())
        if p.returncode != 0:
            printf(p.stderr.read())
        return p.returncode


def choose_file(path_templates, name):
    """
    return the first exists path, if not one exists, return None
    ````
    choose_file(["/bin/?", "/usr/local/?"], "sh")
    ````
    """
    for p in map(lambda pt: pt.replace("?", name), path_templates):
        path = Path(p)
        if path.exists():
            return str(path)
    return None

def get_file_hash(path, update, block_size=65536):
    """
    Read blocks from file of the `path` and call `update` once a block.
    `block_size` is the size of one block, is 65535 bytes (a.k.a. 64 kilobytes).
    """
    with open(path, 'rb') as f:
        fblk = f.read(block_size)
        while len(fblk) > 0:
            update(fblk)
            fblk = f.read(block_size)

def get_file_hash_sha224(path, block_size=65536):
    """
    Get the sha224 hash of file of the `path`.
    Return `str`
    """
    obj = sha224()
    get_file_hash(path, obj.update, block_size)
    return obj.hexdigest()

def check_binary(name, PATH=None):
    """
    check if the given name exists in PATHs, if exists, return the first path found, else return None.
    ````
    if check_binary("clang"):
        print("Clang mounted!")

    if not check_binary("some-tool", PATH=".:./tools/"):
        print("Could not found the tool needed for compiling")
    ````
    """
    PATH = PATH if PATH else os.environ.get("PATH", "")
    pathes = map(lambda x: os.path.join(x, "?"), PATH.split(":"))
    return choose_file(pathes, name)

def create_makru(*args):
    """
    create a makru instance to deal with another makru project. this function doesn't have path expanding, so the '.' means CWD, not the folder contains makru config file.
    ````
    projectb = create_makru("-F","./projectb")
    projectb.action("test")
    ````
    """
    from .makru import Makru
    return Makru(args)

def run_makru(*args):
    """
    it looks like run the makru CLI directly.
    """
    makru_instance = create_makru(**args)
    makru_instance.main()


def compare_version(v1, v2):
    """
    This function is a warpper of semver.VersionInfo.compare().
    Return `int`. The return value is negative if v1 < v2, zero if v1 == v2 and strictly positive if v1 > v2.
    """
    return semver.VersionInfo.parse(v1).compare(v2)


def match_version(v1, cond):
    """
    This function is a warpper of semver.VersionInfo.match().
    `cond` is a `str` prefixed with one of > (greater than), < (less than), >= (not less than), <= (not greater than), == (equal), != (not equal).
    Return `bool`.
    """
    return semver.VersionInfo.parse(v1).match(cond)

def is_semver(ver):
    """
    This function is a warpper of semver.VersionInfo.isvalid().
    Return `bool`.
    """
    return semver.VersionInfo.isvalid(ver)


def is_prerelease(ver):
    """
    Return `True` if `ver` contains "prerelease" field (e.g. "pre.2" or "beta.1"), otherwise return `False`.
    """
    return semver.VersionInfo.parse(ver).prerelease != None


def pick_latest_release(vers: Iterable[str]) -> Union[str, None]:
    """
    Return the most recent released version from `vers`. The "released version" means that it does not contains prerelease field.
    Return `str`, or `None` if `vers` is empty.
    """
    iterator_vers = iter(vers)
    latest = next(iterator_vers, None)
    for v in iterator_vers:
        if (not is_prerelease(v)) and (compare_version(latest, v) < 0):
            latest = v
    return latest


def pick_max_version(vers: Iterable[str]) -> Union[str, None]:
    """
    Return the max version from `vers`.
    Return `str`, or `None` if `vers` is empty.
    """
    iterator_vers = iter(vers)
    latest = next(iterator_vers, None)
    for v in iterator_vers:
        if compare_version(latest, v) <= 0:
            latest = v
    return latest
